@@ -21,6 +21,7 @@
 
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 
+#include "util.h"
 #include "zlibsupport.h"
 #include "depmod.h"
 #include "logging.h"
@@ -55,9 +56,9 @@ struct module_search
 
 	/* search path */
 	char *search_path;
+	size_t len;
 };
 
-static int verbose;
 static unsigned int skipchars;
 static unsigned int make_map_files = 1; /* default to on */
 static unsigned int force_map_files = 0; /* default to on */
@@ -156,7 +157,7 @@ static void load_system_map(const char *filename)
 			continue;
 
 		/* Covers gpl-only and normal symbols. */
-		if (strncmp(ptr+1, ksymstr, ksymstr_len) == 0)
+		if (strstarts(ptr+1, ksymstr))
 			add_symbol(ptr+1+ksymstr_len, NULL);
 	}
 
@@ -213,39 +214,6 @@ static int old_module_version(const char *version)
 	return 1;
 }
 
-static void exec_old_depmod(char *argv[])
-{
-	char *sep;
-	char pathname[strlen(argv[0])+1];
-	char oldname[strlen("depmod") + strlen(argv[0]) + sizeof(".old")];
-
-	memset(pathname, 0, strlen(argv[0])+1);
-	sep = strrchr(argv[0], '/');
-	if (sep)
-		memcpy(pathname, argv[0], sep - argv[0]+1);
-	sprintf(oldname, "%s%s.old", pathname, "depmod");
-
-	/* Recursion detection: we need an env var since we can't
-	   change argv[0] (as older modutils uses it to determine
-	   behavior). */
-	if (getenv("MODULE_RECURSE"))
-		return;
-	setenv("MODULE_RECURSE", "y", 0);
-
-	execvp(oldname, argv);
-	fprintf(stderr,
-		"Version requires old depmod, but couldn't run %s: %s\n",
-		oldname, strerror(errno));
-	exit(2);
-}
-
-static void grammar(const char *cmd, const char *filename, unsigned int line)
-{
-	warn("%s line %u: ignoring bad line starting with '%s'\n",
-						filename, line, cmd);
-}
-
-
 static void print_usage(const char *name)
 {
 	fprintf(stderr,
@@ -325,6 +293,7 @@ static struct module *grab_module(const char *dirname, const char *filename)
 	new->basename = my_basename(new->pathname);
 
 	INIT_LIST_HEAD(&new->dep_list);
+	new->order = INDEX_PRIORITY_MIN;
 
 	new->data = grab_file(new->pathname, &new->len);
 	if (!new->data) {
@@ -455,26 +424,14 @@ static void del_module(struct module **modules, struct module *delme)
 	deleted = delme;
 }
 
-/* Convert filename to the module name.  Works if filename == modname, too. */
-static void filename2modname(char *modname, const char *filename)
+/* convert to relative path if possible */
+static const char *compress_path(const char *path, const char *basedir)
 {
-	const char *afterslash;
-	unsigned int i;
+	int len = strlen(basedir);
 
-	afterslash = strrchr(filename, '/');
-	if (!afterslash)
-		afterslash = filename;
-	else
-		afterslash++;
-
-	/* Convert to underscores, stop at first . */
-	for (i = 0; afterslash[i] && afterslash[i] != '.'; i++) {
-		if (afterslash[i] == '-')
-			modname[i] = '_';
-		else
-			modname[i] = afterslash[i];
-	}
-	modname[i] = '\0';
+	if (strncmp(path, basedir, len) == 0)
+		path += len + 1;
+	return path;
 }
 
 static void output_deps(struct module *modules,
@@ -486,11 +443,12 @@ static void output_deps(struct module *modules,
 		struct list_head *j, *tmp;
 		order_dep_list(i, i);
 
-		fprintf(out, "%s:", i->pathname + strlen(dirname)+1);
+		fprintf(out, "%s:", compress_path(i->pathname, dirname));
 		list_for_each_safe(j, tmp, &i->dep_list) {
 			struct module *dep
 				= list_entry(j, struct module, dep_list);
-			fprintf(out, " %s", dep->pathname + strlen(dirname)+1);
+			fprintf(out, " %s",
+			        compress_path(dep->pathname, dirname));
 			list_del_init(j);
 		}
 		fprintf(out, "\n");
@@ -516,18 +474,21 @@ static void output_deps_bin(struct module *modules,
 		
 		order_dep_list(i, i);
 		
-		filename2modname(modname, i->pathname + strlen(dirname)+1);
-		nofail_asprintf(&line, "%s %s:", modname, i->pathname + strlen(dirname)+1);
+		filename2modname(modname, i->pathname);
+		nofail_asprintf(&line, "%s:",
+		                compress_path(i->pathname, dirname));
 		p = line;
 		list_for_each_safe(j, tmp, &i->dep_list) {
 			struct module *dep
 				= list_entry(j, struct module, dep_list);
-			nofail_asprintf(&line, "%s %s", p, dep->pathname + strlen(dirname)+1);
+			nofail_asprintf(&line, "%s %s",
+			                p,
+			                compress_path(dep->pathname, dirname));
 			free(p);
 			p = line;
 			list_del_init(j);
 		}
-		if (index_insert(index, line) && warn_dups)
+		if (index_insert(index, modname, line, i->order) && warn_dups)
 			warn("duplicate module deps:\n%s\n",line);
 		free(line);
 	}
@@ -552,7 +513,6 @@ static int is_higher_priority(const char *newpath, const char *oldpath,
 			      struct module_search *search,
 			      struct module_overrides *overrides)
 {
-	char *p, *q, *s;
 	struct module_search *tmp;
 	struct module_overrides *ovtmp;
 	int i = 0;
@@ -560,63 +520,29 @@ static int is_higher_priority(const char *newpath, const char *oldpath,
 	int prio_new = -1;
 	int prio_old = -1;
 
-/* The names already match, now we check for wildcard overrides and other
- * high priority overrides we added to the config.
+/* The names already match, now we check for overrides and directory search
+ * order
  */
-	for (ovtmp=overrides;ovtmp!=NULL;ovtmp=ovtmp->next) {
-
-		p = strstr(ovtmp->modfile,newpath);
-		if (p)
+	for (ovtmp = overrides; ovtmp != NULL; ovtmp = ovtmp->next) {
+		if (streq(ovtmp->modfile, newpath))
 			return 1;
-
-		q = strstr(ovtmp->modfile,"*");
-		if (q) {
-			p = strstr(newpath, q+1);
-			if (p)
-				return 1;
-
-			p = strstr(oldpath, q+1);
-			if (p)
-				return 0;
-		}
-
-		p = strstr(ovtmp->modfile,oldpath);
-		if (p)
+		if (streq(ovtmp->modfile, oldpath))
 			return 0;
 	}
-	
-	for (i=0,tmp=search;tmp!=NULL;tmp=tmp->next,i++) {
-
-		s = NOFAIL(malloc(strlen(tmp->search_path)+2));
-		s[0] = '/';
-		strncpy(s+1,tmp->search_path, strlen(tmp->search_path)+1);
-
-		if (0 == strncmp(tmp->search_path,MODULE_BUILTIN_KEY,
-				 strlen(MODULE_BUILTIN_KEY)))
+	for (i = 0, tmp = search; tmp != NULL; tmp = tmp->next, i++) {
+		if (streq(tmp->search_path, MODULE_BUILTIN_KEY))
 			prio_builtin = i;
-			
-		p = strstr(newpath,s);
-		if ((p) && ((p[strlen(s)] == '/')
-			||  (p[strlen(s)] == '\0')))
+		else if (strncmp(tmp->search_path, newpath, tmp->len) == 0)
 			prio_new = i;
-
-		p = strstr(oldpath,s);
-		if ((p) && ((p[strlen(s)] == '/')
-			||  (p[strlen(s)] == '\0')))
+		else if (strncmp(tmp->search_path, oldpath, tmp->len) == 0)
 			prio_old = i;
-	
-		free(s);
-		
 	}
-
 	if (prio_new < 0)
 		prio_new = prio_builtin;
-	
 	if (prio_old < 0)
 		prio_old = prio_builtin;
 
 	return prio_new > prio_old;
-	
 }
 
 
@@ -718,6 +644,7 @@ static struct module *sort_modules(const char *dirname, struct module *list)
 	int dir_len = strlen(dirname) + 1;
 	char file_name[dir_len + strlen("modules.order") + 1];
 	char line[10240];
+	unsigned int linenum = 0;
 
 	sprintf(file_name, "%s/%s", dirname, "modules.order");
 
@@ -737,11 +664,13 @@ static struct module *sort_modules(const char *dirname, struct module *list)
 		struct module **pos, *mod;
 		int len = strlen(line);
 
+		linenum++;
 		if (line[len - 1] == '\n')
 			line[len - 1] = '\0';
 
 		for (pos = &list; (mod = *pos); pos = &(*pos)->next) {
-			if (strcmp(line, mod->pathname + dir_len) == 0) {
+			if (streq(line, mod->pathname + dir_len)) {
+				mod->order = linenum;
 				*pos = mod->next;
 				mod->next = NULL;
 				*tpos = mod;
@@ -769,7 +698,7 @@ static struct module *parse_modules(struct module *list)
 	}
 	
 	for (i = list; i; i = i->next)
-		i->ops->calculate_deps(i, verbose);
+		i->ops->calculate_deps(i);
 	
 	/* Strip out modules with dependency loops. */
  again:
@@ -809,7 +738,8 @@ static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 {
 	struct index_node *index;
 	unsigned int i;
-	char *line;
+	char *alias;
+	int duplicate;
 
 	index = index_create();
 	
@@ -820,12 +750,13 @@ static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 			if (s->owner) {
 				char modname[strlen(s->owner->pathname)+1];
 				filename2modname(modname, s->owner->pathname);
-				nofail_asprintf(&line, "symbol:%s %s",
-					s->name, modname);
-				if (index_insert(index, line) && warn_dups)
-					warn("duplicate module syms:\n%s\n",
-						line);
-				free(line);
+				nofail_asprintf(&alias, "symbol:%s", s->name);
+				duplicate = index_insert(index, alias, modname,
+							 s->owner->order);
+				if (duplicate && warn_dups)
+					warn("duplicate module syms:\n%s %s\n",
+						alias, modname);
+				free(alias);
 			}
 		}
 	}
@@ -834,47 +765,6 @@ static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 	index_destroy(index);
 }
 
-static const char *next_string(const char *string, unsigned long *secsize)
-{
-	/* Skip non-zero chars */
-	while (string[0]) {
-		string++;
-		if ((*secsize)-- <= 1)
-			return NULL;
-	}
-
-	/* Skip any zero padding. */
-	while (!string[0]) {
-		string++;
-		if ((*secsize)-- <= 1)
-			return NULL;
-	}
-	return string;
-}
-
-/* Careful!  Don't munge - in [ ] as per Debian Bug#350915 */
-static char *underscores(char *string)
-{
-	unsigned int i;
-
-	if (!string)
-		return NULL;
-		
-	for (i = 0; string[i]; i++) {
-		switch (string[i]) {
-		case '-':
-			string[i] = '_';
-			break;
-		
-		case '[':
-			i += strcspn(&string[i], "]");
-			if (!string[i])
-				warn("Unmatched bracket in %s\n", string);
-		}
-	}
-	return string;
-}
-
 static void output_aliases(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
@@ -897,7 +787,7 @@ static void output_aliases(struct module *modules, FILE *out, char *dirname)
 		for (p = i->ops->get_modinfo(i, &size);
 		     p;
 		     p = next_string(p, &size)) {
-			if (strncmp(p, "alias=", strlen("alias=")) == 0)
+			if (strstarts(p, "alias="))
 				fprintf(out, "alias %s %s\n",
 					p + strlen("alias="), modname);
 		}
@@ -908,9 +798,10 @@ static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
 	const char *p;
-	char *line;
+	char *alias;
 	unsigned long size;
 	struct index_node *index;
+	int duplicate;
 
 	index = index_create();
 	
@@ -923,25 +814,27 @@ static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 		for (p = i->ops->get_aliases(i, &size);
 		     p;
 		     p = next_string(p, &size)) {
-			nofail_asprintf(&line, "%s %s", p, modname);
-			underscores(line);
-			if (index_insert(index, line) && warn_dups)
-				warn("duplicate module alias:\n%s\n", line);
-			free(line);
+			alias = NOFAIL(strdup(p));
+			underscores(alias);
+			duplicate = index_insert(index, alias, modname, i->order);
+			if (duplicate && warn_dups)
+				warn("duplicate module alias:\n%s %s\n",
+					alias, modname);
+			free(alias);
 		}
 
 		/* Grab from new-style .modinfo section. */
 		for (p = i->ops->get_modinfo(i, &size);
 		     p;
 		     p = next_string(p, &size)) {
-			if (strncmp(p, "alias=", strlen("alias=")) == 0) {
-				nofail_asprintf(&line, "%s %s",
-					p + strlen("alias="), modname);
-				underscores(line);
-				if (index_insert(index, line) && warn_dups)
-					warn("duplicate module alias:\n%s\n",
-						line);
-				free(line);
+			if (strstarts(p, "alias=")) {
+				alias = NOFAIL(strdup(p + strlen("alias=")));
+				underscores(alias);
+				duplicate = index_insert(index, alias, modname, i->order);
+				if (duplicate && warn_dups)
+					warn("duplicate module alias:\n%s %s\n",
+						alias, modname);
+				free(alias);
 			}
 		}
 	}
@@ -1023,52 +916,6 @@ static int depfile_out_of_date(const char *dirname)
 	return any_modules_newer(dirname, st.st_mtime);
 }
 
-static char *getline_wrapped(FILE *file, unsigned int *linenum)
-{
-	int size = 256;
-	int i = 0;
-	char *buf = NOFAIL(malloc(size));
-	for(;;) {
-		int ch = getc_unlocked(file);
-		
-		switch(ch) {
-		case EOF:
-			if (i == 0) {
-				free(buf);
-				return NULL;
-			}
-			/* else fall through */
-			
-		case '\n':
-			if (linenum)
-				(*linenum)++;
-			if (i == size)
-				buf = NOFAIL(realloc(buf, size + 1));
-			buf[i] = '\0';
-			return buf;
-			
-		case '\\':
-			ch = getc_unlocked(file);
-			
-			if (ch == '\n') {
-				if (linenum)
-					(*linenum)++;
-				continue;
-			}
-			/* else fall through */
-		
-		default:
-			buf[i++] = ch;
-	
-			if (i == size) {
-				size *= 2;
-				buf = NOFAIL(realloc(buf, size));
-			}
-		}
-	}
-}
-
-
 static char *strsep_skipspace(char **string, char *delim)
 {
 	if (!*string)
@@ -1078,6 +925,7 @@ static char *strsep_skipspace(char **string, char *delim)
 }
 
 static struct module_search *add_search(const char *search_path,
+					size_t len,
 					struct module_search *search)
 {
 
@@ -1085,6 +933,7 @@ static struct module_search *add_search(const char *search_path,
 	
 	new = NOFAIL(malloc(sizeof(*new)));
 	new->search_path = NOFAIL(strdup(search_path));
+	new->len = len;
 	new->next = search;
 
 	return new;
@@ -1105,16 +954,17 @@ static struct module_overrides *add_override(const char *modfile,
 	
 }
 
-/* Recursion */
-static int read_config(const char *filename,
-		       const char *basedir,
-		       struct module_search **search,
-		       struct module_overrides **overrides);
+static int parse_config_scan(const char *filename,
+			     const char *basedir,
+			     const char *kernelversion,
+			     struct module_search **search,
+			     struct module_overrides **overrides);
 
-static int read_config_file(const char *filename,
-			    const char *basedir,
-			    struct module_search **search,
-                            struct module_overrides **overrides)
+static int parse_config_file(const char *filename,
+			     const char *basedir,
+			     const char *kernelversion,
+			     struct module_search **search,
+			     struct module_overrides **overrides)
 {
 	char *line;
 	unsigned int linenum = 0;
@@ -1139,52 +989,71 @@ static int read_config_file(const char *filename,
 			continue;
 		}
 
-		if (strcmp(cmd, "search") == 0) {
+		if (streq(cmd, "search")) {
 			char *search_path;
 			
-			while ((search_path = strsep_skipspace(&ptr, "\t ")))
-				*search = add_search(search_path, *search);
-		} else if (strcmp(cmd, "override") == 0) {
+			while ((search_path = strsep_skipspace(&ptr, "\t "))) {
+				char *dirname;
+				size_t len;
+
+				if (strcmp(search_path,
+						MODULE_BUILTIN_KEY) == 0) {
+					*search = add_search(MODULE_BUILTIN_KEY,
+							     0, *search);
+					continue;
+				}
+				nofail_asprintf(&dirname, "%s%s%s/%s", basedir,
+					MODULE_DIR, kernelversion, search_path);
+				len = strlen(dirname);
+				*search = add_search(dirname, len, *search);
+				free(dirname);
+			}
+		} else if (streq(cmd, "override")) {
 			char *pathname = NULL, *version, *subdir;
 			modname = strsep_skipspace(&ptr, "\t ");
 			version = strsep_skipspace(&ptr, "\t ");
 			subdir = strsep_skipspace(&ptr, "\t ");
 
-			pathname = NOFAIL(malloc(strlen(basedir)
-				               + strlen(MODULE_DIR)
-					       + strlen(version)
-					       + strlen(subdir)
-					       + strlen(modname)
-					       + strlen(".ko")
-					       + 3));
-			sprintf(pathname, "%s%s%s/%s/%s.ko", basedir,
-				MODULE_DIR, version, subdir, modname);
+			if (strcmp(version, kernelversion) != 0 &&
+			    strcmp(version, "*") != 0)
+				continue;
+
+			nofail_asprintf(&pathname, "%s%s%s/%s/%s.ko", basedir,
+				MODULE_DIR, kernelversion, subdir, modname);
 
 			*overrides = add_override(pathname, *overrides);
-		} else if (strcmp(cmd, "include") == 0) {
+			free(pathname);
+		} else if (streq(cmd, "include")) {
 			char *newfilename;
 
 			newfilename = strsep_skipspace(&ptr, "\t ");
-			if (!newfilename)
+			if (!newfilename) {
 				grammar(cmd, filename, linenum);
-			else {
-				if (!read_config(newfilename, basedir,
-						 search, overrides))
+			} else {
+				warn("\"include %s\" is deprecated, "
+				     "please use /etc/depmod.d\n", newfilename);
+				if (strstarts(newfilename, "/etc/depmod.d")) {
+					warn("\"include /etc/depmod.d\" is "
+					     "the default, ignored\n");
+				} else {
+					if (!parse_config_scan(newfilename, basedir,
+							       kernelversion,
+							       search, overrides))
 					warn("Failed to open included"
 					     " config file %s: %s\n",
 					     newfilename, strerror(errno));
-                        }
-
-		} else if (strcmp(cmd, "make_map_files") == 0) {
+				}
+			}
+		} else if (streq(cmd, "make_map_files")) {
 			char *option;
 
 			option = strsep_skipspace(&ptr, "\t ");
 			if (!option)
 				grammar(cmd, filename, linenum);
 			else {
-				if (0 == strncmp(option, "yes", 3))
+				if (streq(option, "yes"))
 					make_map_files = 1;
-				else if (0 == strncmp(option, "no", 2))
+				else if (streq(option, "no"))
 					make_map_files = 0;
 				else
 					grammar(cmd, filename, linenum);
@@ -1198,67 +1067,97 @@ static int read_config_file(const char *filename,
         return 1;
 }
 
-/* Simple format, ignore lines starting with #, one command per line.
-   Returns true or false. */
-static int read_config(const char *filename,
-		       const char *basedir,
-		       struct module_search **search,
-                       struct module_overrides **overrides)
+static int parse_config_scan(const char *filename,
+			     const char *basedir,
+			     const char *kernelversion,
+			     struct module_search **search,
+			     struct module_overrides **overrides)
 {
 	DIR *dir;
 	int ret = 0;
 
 	dir = opendir(filename);
 	if (dir) {
+		struct file_entry {
+			struct list_head node;
+			char name[];
+		};
+		LIST_HEAD(files_list);
+		struct file_entry *fe, *fe_tmp;
 		struct dirent *i;
-		while ((i = readdir(dir)) != NULL) {
-			if (!streq(i->d_name,".") && !streq(i->d_name,"..")
-					&& config_filter(i->d_name)) {
-				char sub[strlen(filename) + 1
-				       + strlen(i->d_name) + 1];
 
-				sprintf(sub, "%s/%s", filename, i->d_name);
-				if (!read_config(sub, basedir, search,
-						 overrides))
-					warn("Failed to open"
-					     " config file %s: %s\n",
-					     sub, strerror(errno));
-			}
+		/* sort files from directory into list */
+		while ((i = readdir(dir)) != NULL) {
+			size_t len;
+
+			if (i->d_name[0] == '.')
+				continue;
+			if (!config_filter(i->d_name))
+				continue;
+
+			len = strlen(i->d_name);
+			if (len < 6 || strcmp(&i->d_name[len-5], ".conf") != 0)
+				warn("All config files need .conf: %s/%s, "
+				     "it will be ignored in a future release.\n",
+				     filename, i->d_name);
+			fe = malloc(sizeof(struct file_entry) + len + 1);
+			if (fe == NULL)
+				continue;
+			strcpy(fe->name, i->d_name);
+			list_for_each_entry(fe_tmp, &files_list, node)
+				if (strcmp(fe_tmp->name, fe->name) >= 0)
+					break;
+			list_add_tail(&fe->node, &fe_tmp->node);
 		}
 		closedir(dir);
+
+		/* parse list of files */
+		list_for_each_entry_safe(fe, fe_tmp, &files_list, node) {
+			char *cfgfile;
+
+			nofail_asprintf(&cfgfile, "%s/%s", filename, fe->name);
+			if (!parse_config_file(cfgfile, basedir, kernelversion,
+					       search, overrides))
+				warn("Failed to open config file "
+				     "%s: %s\n", fe->name, strerror(errno));
+			free(cfgfile);
+			list_del(&fe->node);
+			free(fe);
+		}
+
 		ret = 1;
 	} else {
-		if (read_config_file(filename, basedir, search, overrides))
+		if (parse_config_file(filename, basedir, kernelversion, search,
+				      overrides))
 			ret = 1;
 	}
 
 	return ret;
 }
 
-static const char *default_configs[] =
+static void parse_toplevel_config(const char *filename,
+				  const char *basedir,
+				  const char *kernelversion,
+				  struct module_search **search,
+				  struct module_overrides **overrides)
 {
-	"/etc/depmod.conf",
-	"/etc/depmod.d",
-};
-
-static void read_toplevel_config(const char *filename,
-				 const char *basedir,
-				 struct module_search **search,
-                                 struct module_overrides **overrides)
-{
-	unsigned int i;
-
 	if (filename) {
-		if (!read_config(filename, basedir, search, overrides))
+		if (!parse_config_scan(filename, basedir, kernelversion, search,
+				 overrides))
 			fatal("Failed to open config file %s: %s\n",
 			      filename, strerror(errno));
 		return;
 	}
 
-	/* Try defaults. */
-	for (i = 0; i < ARRAY_SIZE(default_configs); i++) {
-		read_config(default_configs[i], basedir, search, overrides);
-	}
+	/* deprecated config file */
+	if (parse_config_file("/etc/depmod.conf", basedir, kernelversion,
+			      search, overrides) > 0)
+		warn("Deprecated config file /etc/depmod.conf, "
+		      "all config files belong into /etc/depmod.d/.\n");
+
+	/* default config */
+	parse_config_scan("/etc/depmod.d", basedir, kernelversion,
+			  search, overrides);
 }
 
 /* Local to main, but not freed on exit.  Keep valgrind quiet. */
@@ -1343,9 +1242,12 @@ int main(int argc, char *argv[])
 		version = NOFAIL(strdup(buf.release));
 	}
 
-	/* Run old version if required. */
-	if (old_module_version(version))
-		exec_old_depmod(argv);
+	/* Check for old version. */
+	if (old_module_version(version)) {
+		fprintf(stderr, "Kernel version %s requires old depmod\n",
+			version);
+		exit(2);
+	}
 
 	if (badopt) {
 		fprintf(stderr, "%s: malformed/unrecognized option '%s'\n",
@@ -1358,10 +1260,7 @@ int main(int argc, char *argv[])
 	if (optind == argc)
 		all = 1;
 
-	dirname = NOFAIL(malloc(strlen(basedir)
-			 + strlen(MODULE_DIR)
-			 + strlen(version) + 1));
-	sprintf(dirname, "%s%s%s", basedir, MODULE_DIR, version);
+	nofail_asprintf(&dirname, "%s%s%s", basedir, MODULE_DIR, version);
 
 	if (maybe_all) {
 		if (!doing_stdout && !depfile_out_of_date(dirname))
@@ -1369,19 +1268,30 @@ int main(int argc, char *argv[])
 		all = 1;
 	}
 
-	read_toplevel_config(config, basedir, &search, &overrides);
+	parse_toplevel_config(config, basedir, version, &search, &overrides);
 
 	/* For backward compatibility add "updates" to the head of the search
 	 * list here. But only if there was no "search" option specified.
 	 */
+	if (!search) {
+		char *dirname;
+		size_t len;
 
-	if (!search)
-		search = add_search("updates",search);
-	
+		nofail_asprintf(&dirname, "%s%s%s/updates", basedir,
+				MODULE_DIR, version);
+		len = strlen(dirname);
+		search = add_search(dirname, len, search);
+	}
 	if (!all) {
 		/* Do command line args. */
 		for (opt = optind; opt < argc; opt++) {
-			struct module *new = grab_module(NULL, argv[opt]);
+			struct module *new;
+
+			if (argv[opt][0] != '/')
+				fatal("modules must be specified using absolute paths.\n"
+					"\"%s\" is a relative path\n", argv[opt]);
+
+			new = grab_module(NULL, argv[opt]);
 			if (!new) {
 				/* cmd-line specified modules must exist */
 				fatal("grab_module() failed for module %s\n", argv[opt]);
