@@ -10,6 +10,15 @@
 /* Do we use syslog for messages or stderr? */
 int logging = 0;
 
+/* Do we want to silently drop all warnings? */
+int quiet = 0;
+
+/* Number of times warn() has been called */
+int warned = 0;
+
+/* Do we want informative messages as well as errors? */
+int verbose = 0;
+
 void message(const char *prefix, const char *fmt, va_list *arglist)
 {
 	int ret;
@@ -34,13 +43,21 @@ void message(const char *prefix, const char *fmt, va_list *arglist)
 	free(buf);
 }
 
-int warned = 0;
 void warn(const char *fmt, ...)
 {
 	va_list arglist;
 	warned++;
 	va_start(arglist, fmt);
-	message("WARNING: ", fmt, &arglist);
+	if (!quiet)
+		message("WARNING: ", fmt, &arglist);
+	va_end(arglist);
+}
+
+void error(const char *fmt, ...)
+{
+	va_list arglist;
+	va_start(arglist, fmt);
+	message("ERROR: ", fmt, &arglist);
 	va_end(arglist);
 }
 
@@ -53,3 +70,14 @@ void fatal(const char *fmt, ...)
 	exit(1);
 }
 
+/* If we don't flush, then child processes print before we do */
+void info(const char *fmt, ...)
+{
+	va_list arglist;
+	va_start(arglist, fmt);
+	if (verbose) {
+		vfprintf(stdout, fmt, arglist);
+		fflush(stdout);
+	}
+	va_end(arglist);
+}
