@@ -39,9 +39,7 @@
 #include <sys/wait.h>
 #include <syslog.h>
 
-#define streq(a,b) (strcmp((a),(b)) == 0)
-#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
-
+#include "util.h"
 #include "zlibsupport.h"
 #include "logging.h"
 #include "index.h"
@@ -67,12 +65,6 @@ struct module {
 
 typedef void (*errfn_t)(const char *fmt, ...);
 
-static void grammar(const char *cmd, const char *filename, unsigned int line)
-{
-	warn("%s line %u: ignoring bad line starting with '%s'\n",
-	     filename, line, cmd);
-}
-
 static void print_usage(const char *progname)
 {
 	fprintf(stderr,
@@ -83,102 +75,30 @@ static void print_usage(const char *progname)
 	exit(1);
 }
 
-static char *getline_wrapped(FILE *file, unsigned int *linenum)
-{
-	int size = 256;
-	int i = 0;
-	char *buf = NOFAIL(malloc(size));
-	for(;;) {
-		int ch = getc_unlocked(file);
-		
-		switch(ch) {
-		case EOF:
-			if (i == 0) {
-				free(buf);
-				return NULL;
-			}
-			/* else fall through */
-			
-		case '\n':
-			if (linenum)
-				(*linenum)++;
-			if (i == size)
-				buf = NOFAIL(realloc(buf, size + 1));
-			buf[i] = '\0';
-			return buf;
-			
-		case '\\':
-			ch = getc_unlocked(file);
-			
-			if (ch == '\n') {
-				if (linenum)
-					(*linenum)++;
-				continue;
-			}
-			/* else fall through */
-		
-		default:
-			buf[i++] = ch;
-	
-			if (i == size) {
-				size *= 2;
-				buf = NOFAIL(realloc(buf, size));
-			}
-		}
-	}
-}
-
 static struct module *find_module(const char *filename, struct list_head *list)
 {
 	struct module *i;
 
 	list_for_each_entry(i, list, list) {
-		if (strcmp(i->filename, filename) == 0)
+		if (streq(i->filename, filename))
 			return i;
 	}
 	return NULL;
 }
 
-/* Convert filename to the module name.  Works if filename == modname, too. */
-static void filename2modname(char *modname, const char *filename)
+/* We used to lock with a write flock but that allows regular users to block
+ * module load by having a read lock on the module file (no way to bust the
+ * existing locks without killing the offending process). Instead, we now
+ * do the system call/init_module and allow the kernel to fail us instead.
+ */
+static int open_file(const char *filename)
 {
-	const char *afterslash;
-	unsigned int i;
+	int fd = open(filename, O_RDONLY, 0);
 
-	afterslash = strrchr(filename, '/');
-	if (!afterslash)
-		afterslash = filename;
-	else
-		afterslash++;
-
-	/* Convert to underscores, stop at first . */
-	for (i = 0; afterslash[i] && afterslash[i] != '.'; i++) {
-		if (afterslash[i] == '-')
-			modname[i] = '_';
-		else
-			modname[i] = afterslash[i];
-	}
-	modname[i] = '\0';
-}
-
-static int lock_file(const char *filename)
-{
-	int fd = open(filename, O_RDWR, 0);
-
-	if (fd >= 0) {
-		struct flock lock;
-		lock.l_type = F_WRLCK;
-		lock.l_whence = SEEK_SET;
-		lock.l_start = 0;
-		lock.l_len = 1;
-		fcntl(fd, F_SETLKW, &lock);
-	} else
-		/* Read-only filesystem?  There goes locking... */
-		fd = open(filename, O_RDONLY, 0);
 	return fd;
 }
 
-static void unlock_file(int fd)
+static void close_file(int fd)
 {
 	/* Valgrind is picky... */
 	close(fd);
@@ -280,16 +200,16 @@ static int add_modules_dep_line(char *line,
 	return 1;
 }
 
-static int read_depends_bin(const char *dirname,
-			 const char *start_name,
-			 struct list_head *list)
+static int read_depends_file(const char *dirname,
+			     const char *start_name,
+			     struct list_head *list)
 {
 	char *modules_dep_name;
 	char *line;
-	FILE *modules_dep;
+	struct index_file *modules_dep;
 
 	nofail_asprintf(&modules_dep_name, "%s/%s", dirname, "modules.dep.bin");
-	modules_dep = fopen(modules_dep_name, "r");
+	modules_dep = index_file_open(modules_dep_name);
 	if (!modules_dep) {
 		free(modules_dep_name);
 		return 0;
@@ -302,8 +222,8 @@ static int read_depends_bin(const char *dirname,
 			fatal("Module index is inconsistent\n");
 		free(line);
 	}
-	
-	fclose(modules_dep);
+
+	index_file_close(modules_dep);
 	free(modules_dep_name);
 	
 	return 1;
@@ -318,8 +238,9 @@ static void read_depends(const char *dirname,
 	FILE *modules_dep;
 	int done = 0;
 
-	if (read_depends_bin(dirname, start_name, list))
-		return;
+	if (use_binary_indexes)
+		if (read_depends_file(dirname, start_name, list))
+			return;
 
 	nofail_asprintf(&modules_dep_name, "%s/%s", dirname, "modules.dep");
 	modules_dep = fopen(modules_dep_name, "r");
@@ -405,7 +326,7 @@ static void *get_section32(void *file,
 		
 	secnames = file + sechdrs[hdr->e_shstrndx].sh_offset;
 	for (i = 1; i < hdr->e_shnum; i++)
-		if (strcmp(secnames + sechdrs[i].sh_name, name) == 0) {
+		if (streq(secnames + sechdrs[i].sh_name, name)) {
 			*secsize = sechdrs[i].sh_size;
 			return file + sechdrs[i].sh_offset;
 		}
@@ -432,7 +353,7 @@ static void *get_section64(void *file,
 		
 	secnames = file + sechdrs[hdr->e_shstrndx].sh_offset;
 	for (i = 1; i < hdr->e_shnum; i++)
-		if (strcmp(secnames + sechdrs[i].sh_name, name) == 0) {
+		if (streq(secnames + sechdrs[i].sh_name, name)) {
 			*secsize = sechdrs[i].sh_size;
 			return file + sechdrs[i].sh_offset;
 		}
@@ -495,7 +416,7 @@ static void invalidate_section32(void *mod, const char *secname)
 	unsigned int i;
 
 	for (i = 1; i < hdr->e_shnum; i++)
-		if (strcmp(secnames+sechdrs[i].sh_name, secname) == 0)
+		if (streq(secnames+sechdrs[i].sh_name, secname))
 			sechdrs[i].sh_flags &= ~SHF_ALLOC;
 }
 
@@ -507,7 +428,7 @@ static void invalidate_section64(void *mod, const char *secname)
 	unsigned int i;
 
 	for (i = 1; i < hdr->e_shnum; i++)
-		if (strcmp(secnames+sechdrs[i].sh_name, secname) == 0)
+		if (streq(secnames+sechdrs[i].sh_name, secname))
 			sechdrs[i].sh_flags &= ~(unsigned long long)SHF_ALLOC;
 }
 
@@ -529,24 +450,6 @@ static void strip_section(struct module *module,
 	}
 }
 
-static const char *next_string(const char *string, unsigned long *secsize)
-{
-	/* Skip non-zero chars */
-	while (string[0]) {
-		string++;
-		if ((*secsize)-- <= 1)
-			return NULL;
-	}
-
-	/* Skip any zero padding. */
-	while (!string[0]) {
-		string++;
-		if ((*secsize)-- <= 1)
-			return NULL;
-	}
-	return string;
-}
-
 static void clear_magic(struct module *module, void *mod, unsigned long len)
 {
 	const char *p;
@@ -559,7 +462,7 @@ static void clear_magic(struct module *module, void *mod, unsigned long len)
 	for (p = get_section(mod, len, ".modinfo", &modlen);
 	     p;
 	     p = next_string(p, &modlen)) {
-		if (strncmp(p, "vermagic=", strlen("vermagic=")) == 0) {
+		if (strstarts(p, "vermagic=")) {
 			memset((char *)p, 0, strlen(p));
 			return;
 		}
@@ -655,7 +558,7 @@ static  int
 find_blacklist(const char *modname, const struct module_blacklist *blacklist)
 {
 	while (blacklist) {
-		if (strcmp(blacklist->modulename, modname) == 0)
+		if (streq(blacklist->modulename, modname))
 			return 1;
 		blacklist = blacklist->next;
 	}
@@ -698,14 +601,34 @@ static char *append_option(char *options, const char *newoption)
 	return options;
 }
 
+static char *prepend_option(char *options, const char *newoption)
+{
+	size_t l1, l2;
+	l1 = strlen(options);
+	l2 = strlen(newoption);
+	/* the resulting string will look like
+	 * newoption + ' ' + options + '\0' */
+	if (l1) {
+		options = NOFAIL(realloc(options, l2 + 1 + l1 + 1));
+		memmove(options + l2 + 1, options, l1 + 1);
+		options[l2] = ' ';
+		memcpy(options, newoption, l2);
+	} else {
+		options = NOFAIL(realloc(options, l2 + 1));
+		memcpy(options, newoption, l2);
+		options[l2] = '\0';
+	}
+	return options;
+}
+
 /* Add to options */
 static char *add_extra_options(const char *modname,
 			       char *optstring,
 			       const struct module_options *options)
 {
 	while (options) {
-		if (strcmp(options->modulename, modname) == 0)
-			optstring = append_option(optstring, options->options);
+		if (streq(options->modulename, modname))
+			optstring = prepend_option(optstring, options->options);
 		options = options->next;
 	}
 	return optstring;
@@ -741,7 +664,11 @@ static int module_in_kernel(const char *modname, unsigned int *usecount)
 	const int ATTR_LEN = 16;
 	char attr[ATTR_LEN];
 
-	/* Find module. We assume sysfs is mounted. */
+	/* Check sysfs is mounted */
+	if (stat("/sys/module", &finfo) < 0)
+		return -1;
+
+	/* Find module. */
 	nofail_asprintf(&name, "/sys/module/%s", modname);
 	ret = stat(name, &finfo);
 	free(name);
@@ -774,23 +701,10 @@ static int module_in_kernel(const char *modname, unsigned int *usecount)
 	return 1;
 }
 
-/* If we don't flush, then child processes print before we do */
-static void verbose_printf(int verbose, const char *fmt, ...)
-{
-	va_list arglist;
-
-	if (verbose) {
-		va_start(arglist, fmt);
-		vprintf(fmt, arglist);
-		fflush(stdout);
-		va_end(arglist);
-	}
-}
-
 /* Do an install/remove command: replace $CMDLINE_OPTS if it's specified. */
 static void do_command(const char *modname,
 		       const char *command,
-		       int verbose, int dry_run,
+		       int dry_run,
 		       errfn_t error,
 		       const char *type,
 		       const char *cmdline_opts)
@@ -807,7 +721,7 @@ static void do_command(const char *modname,
 		replaced_cmd = new;
 	}
 
-	verbose_printf(verbose, "%s %s\n", type, replaced_cmd);
+	info("%s %s\n", type, replaced_cmd);
 	if (dry_run)
 		return;
 
@@ -819,13 +733,12 @@ static void do_command(const char *modname,
 }
 
 /* Actually do the insert.  Frees second arg. */
-static void insmod(struct list_head *list,
+static int insmod(struct list_head *list,
 		   char *optstring,
 		   const char *newname,
 		   int first_time,
 		   errfn_t error,
 		   int dry_run,
-		   int verbose,
 		   const struct module_options *options,
 		   const struct module_command *commands,
 		   int ignore_commands,
@@ -839,20 +752,24 @@ static void insmod(struct list_head *list,
 	void *map;
 	const char *command;
 	struct module *mod = list_entry(list->next, struct module, list);
+	int rc = 0;
 
 	/* Take us off the list. */
 	list_del(&mod->list);
 
-	/* Do things we (or parent) depend on first, but don't die if
-	 * they fail. */
+	/* Do things we (or parent) depend on first. */
 	if (!list_empty(list)) {
-		insmod(list, NOFAIL(strdup("")), NULL, 0, warn,
-		       dry_run, verbose, options, commands, 0, ignore_proc,
-		       strip_vermagic, strip_modversion, "");
+		if ((rc = insmod(list, NOFAIL(strdup("")), NULL, 0, warn,
+		       dry_run, options, commands, 0, ignore_proc,
+		       strip_vermagic, strip_modversion, "")) != 0) {
+			error("Error inserting %s (%s): %s\n",
+				mod->modname, mod->filename,
+				insert_moderror(errno));
+			goto out_optstring;
+		}
 	}
 
-	/* Lock before we look, in case it's initializing. */
-	fd = lock_file(mod->filename);
+	fd = open_file(mod->filename);
 	if (fd < 0) {
 		error("Could not open '%s': %s\n",
 		      mod->filename, strerror(errno));
@@ -870,9 +787,8 @@ static void insmod(struct list_head *list,
 
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
-		/* It might recurse: unlock. */
-		unlock_file(fd);
-		do_command(mod->modname, command, verbose, dry_run, error,
+		close_file(fd);
+		do_command(mod->modname, command, dry_run, error,
 			   "install", cmdline_opts);
 		goto out_optstring;
 	}
@@ -896,7 +812,7 @@ static void insmod(struct list_head *list,
 	/* Config file might have given more options */
 	optstring = add_extra_options(mod->modname, optstring, options);
 
-	verbose_printf(verbose, "insmod %s %s\n", mod->filename, optstring);
+	info("insmod %s %s\n", mod->filename, optstring);
 
 	if (dry_run)
 		goto out;
@@ -915,14 +831,15 @@ static void insmod(struct list_head *list,
 			error("Error inserting %s (%s): %s\n",
 			      mod->modname, mod->filename,
 			      insert_moderror(errno));
+			rc = 1;
 	}
  out:
 	release_file(map, len);
  out_unlock:
-	unlock_file(fd);
+	close_file(fd);
  out_optstring:
 	free(optstring);
-	return;
+	return rc;
 }
 
 /* Do recursive removal. */
@@ -931,7 +848,6 @@ static void rmmod(struct list_head *list,
 		  int first_time,
 		  errfn_t error,
 		  int dry_run,
-		  int verbose,
 		  struct module_command *commands,
 		  int ignore_commands,
 		  int ignore_inuse,
@@ -940,26 +856,20 @@ static void rmmod(struct list_head *list,
 {
 	const char *command;
 	unsigned int usecount = 0;
-	int lock;
 	struct module *mod = list_entry(list->next, struct module, list);
 
 	/* Take first one off the list. */
 	list_del(&mod->list);
 
-	/* Ignore failure; it's best effort here. */
-	lock = lock_file(mod->filename);
-
 	if (!name)
 		name = mod->modname;
 
 	/* Even if renamed, find commands to orig. name. */
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
-		/* It might recurse: unlock. */
-		unlock_file(lock);
-		do_command(mod->modname, command, verbose, dry_run, error,
+		do_command(mod->modname, command, dry_run, error,
 			   "remove", cmdline_opts);
-		goto remove_rest_no_unlock;
+		goto remove_rest;
 	}
 
 	if (module_in_kernel(name, &usecount) == 0)
@@ -971,7 +881,7 @@ static void rmmod(struct list_head *list,
 		goto remove_rest;
 	}
 
-	verbose_printf(verbose, "rmmod %s\n", mod->filename);
+	info("rmmod %s\n", mod->filename);
 
 	if (dry_run)
 		goto remove_rest;
@@ -985,11 +895,9 @@ static void rmmod(struct list_head *list,
 	}
 
  remove_rest:
-	unlock_file(lock);
- remove_rest_no_unlock:
 	/* Now do things we depend. */
 	if (!list_empty(list))
-		rmmod(list, NULL, 0, warn, dry_run, verbose, commands,
+		rmmod(list, NULL, 0, warn, dry_run, commands,
 		      0, 1, "", flags);
 	return;
 
@@ -1073,30 +981,6 @@ static int type_matches(const char *path, const char *subpath)
 	return ret;
 }
 
-/* Careful!  Don't munge - in [ ] as per Debian Bug#350915 */
-static char *underscores(char *string)
-{
-	if (string) {
-		unsigned int i;
-		int inbracket = 0;
-		for (i = 0; string[i]; i++) {
-			switch (string[i]) {
-			case '[':
-				inbracket++;
-				break;
-			case ']':
-				inbracket--;
-				break;
-			case '-':
-				if (!inbracket)
-					string[i] = '_';
-			}
-		}
-		if (inbracket)
-			warn("Unmatched bracket in %s\n", string);
-	}
-	return string;
-}
 
 static int do_wildcard(const char *dirname,
 		       const char *type,
@@ -1150,18 +1034,16 @@ static char *strsep_skipspace(char **string, char *delim)
 	return strsep(string, delim);
 }
 
-/* Recursion */
-static int read_config(const char *filename,
-		       const char *name,
-		       int dump_only,
-		       int removing,
-		       struct module_options **options,
-		       struct module_command **commands,
-		       struct module_alias **alias,
-		       struct module_blacklist **blacklist);
+static int parse_config_scan(const char *filename,
+			     const char *name,
+			     int dump_only,
+			     int removing,
+			     struct module_options **options,
+			     struct module_command **commands,
+			     struct module_alias **alias,
+			     struct module_blacklist **blacklist);
 
-/* FIXME: Maybe should be extended to "alias a b [and|or c]...". --RR */
-static int read_config_file(const char *filename,
+static int parse_config_file(const char *filename,
 			    const char *name,
 			    int dump_only,
 			    int removing,
@@ -1191,38 +1073,42 @@ static int read_config_file(const char *filename,
 			continue;
 		}
 
-		if (strcmp(cmd, "alias") == 0) {
-			char *wildcard
-				= underscores(strsep_skipspace(&ptr, "\t "));
-			char *realname
-				= underscores(strsep_skipspace(&ptr, "\t "));
+		if (streq(cmd, "alias")) {
+			char *wildcard = strsep_skipspace(&ptr, "\t ");
+			char *realname = strsep_skipspace(&ptr, "\t ");
 
 			if (!wildcard || !realname)
 				grammar(cmd, filename, linenum);
-			else if (fnmatch(wildcard,name,0) == 0)
-				*aliases = add_alias(realname, *aliases);
-		} else if (strcmp(cmd, "include") == 0) {
+			else if (fnmatch(underscores(wildcard),name,0) == 0)
+				*aliases = add_alias(underscores(realname), *aliases);
+		} else if (streq(cmd, "include")) {
 			struct module_alias *newalias = NULL;
 			char *newfilename;
 
 			newfilename = strsep_skipspace(&ptr, "\t ");
-			if (!newfilename)
+			if (!newfilename) {
 				grammar(cmd, filename, linenum);
-			else {
-				if (!read_config(newfilename, name,
-						 dump_only, removing,
-						 options, commands, &newalias,
-						 blacklist))
-					warn("Failed to open included"
-					      " config file %s: %s\n",
-					      newfilename, strerror(errno));
-
+			} else {
+				warn("\"include %s\" is deprecated, "
+				     "please use /etc/modprobe.d\n", newfilename);
+				if (strstarts(newfilename, "/etc/modprobe.d")) {
+					warn("\"include /etc/modprobe.d\" is "
+					     "the default, ignored\n");
+				} else {
+					if (!parse_config_scan(newfilename, name,
+							      dump_only, removing,
+							      options, commands, &newalias,
+							      blacklist))
+						warn("Failed to open included"
+						      " config file %s: %s\n",
+						      newfilename, strerror(errno));
+				}
 				/* Files included override aliases,
 				   etc that was already set ... */
 				if (newalias)
 					*aliases = newalias;
 			}
-		} else if (strcmp(cmd, "options") == 0) {
+		} else if (streq(cmd, "options")) {
 			modname = strsep_skipspace(&ptr, "\t ");
 			if (!modname || !ptr)
 				grammar(cmd, filename, linenum);
@@ -1231,7 +1117,7 @@ static int read_config_file(const char *filename,
 				*options = add_options(underscores(modname),
 						       ptr, *options);
 			}
-		} else if (strcmp(cmd, "install") == 0) {
+		} else if (streq(cmd, "install")) {
 			modname = strsep_skipspace(&ptr, "\t ");
 			if (!modname || !ptr)
 				grammar(cmd, filename, linenum);
@@ -1240,7 +1126,7 @@ static int read_config_file(const char *filename,
 				*commands = add_command(underscores(modname),
 							ptr, *commands);
 			}
-		} else if (strcmp(cmd, "blacklist") == 0) {
+		} else if (streq(cmd, "blacklist")) {
 			modname = strsep_skipspace(&ptr, "\t ");
 			if (!modname)
 				grammar(cmd, filename, linenum);
@@ -1248,7 +1134,7 @@ static int read_config_file(const char *filename,
 				*blacklist = add_blacklist(underscores(modname),
 							*blacklist);
 			}
-		} else if (strcmp(cmd, "remove") == 0) {
+		} else if (streq(cmd, "remove")) {
 			modname = strsep_skipspace(&ptr, "\t ");
 			if (!modname || !ptr)
 				grammar(cmd, filename, linenum);
@@ -1257,13 +1143,16 @@ static int read_config_file(const char *filename,
 				*commands = add_command(underscores(modname),
 							ptr, *commands);
 			}
-		} else if (strcmp(cmd, "config") == 0) {
+		} else if (streq(cmd, "config")) {
 			char *tmp = strsep_skipspace(&ptr, "\t ");
-			if (strcmp(tmp, "binary_indexes") == 0) {
+
+			if (!tmp)
+				grammar(cmd, filename, linenum);
+			else if (streq(tmp, "binary_indexes")) {
 				tmp = strsep_skipspace(&ptr, "\t ");
-				if (strcmp(tmp, "yes") == 0)
+				if (streq(tmp, "yes"))
 					use_binary_indexes = 1;
-				if (strcmp(tmp, "no") == 0)
+				if (streq(tmp, "no"))
 					use_binary_indexes = 0;
 			}
 		} else
@@ -1275,129 +1164,157 @@ static int read_config_file(const char *filename,
 	return 1;
 }
 
-/* Simple format, ignore lines starting with #, one command per line.
-   Returns true or false. */
-static int read_config(const char *filename,
-		       const char *name,
-		       int dump_only,
-		       int removing,
-		       struct module_options **options,
-		       struct module_command **commands,
-		       struct module_alias **aliases,
-		       struct module_blacklist **blacklist)
+/* fallback to plain-text aliases file as necessary */
+static int read_aliases_file(const char *filename,
+			     const char *name,
+			     int dump_only,
+			     int removing,
+			     struct module_options **options,
+			     struct module_command **commands,
+			     struct module_alias **aliases,
+			     struct module_blacklist **blacklist)
+{
+	struct index_value *realnames;
+	struct index_value *realname;
+	char *binfile;
+	struct index_file *index;
+
+	if (!use_binary_indexes)
+		goto plain_text;
+
+	nofail_asprintf(&binfile, "%s.bin", filename);
+	index = index_file_open(binfile);
+	if (!index) {
+		free(binfile);
+		goto plain_text;
+	}
+
+	if (dump_only) {
+		index_dump(index, stdout, "alias ");
+		free(binfile);
+		index_file_close(index);
+		return 1;
+	}
+
+	realnames = index_searchwild(index, name);
+	for (realname = realnames; realname; realname = realname->next)
+		*aliases = add_alias(realname->value, *aliases);
+	index_values_free(realnames);
+
+	free(binfile);
+	index_file_close(index);
+	return 1;
+
+plain_text:
+	return parse_config_file(filename, name, dump_only, removing,
+				 options, commands, aliases, blacklist);
+}
+
+static int parse_config_scan(const char *filename,
+			     const char *name,
+			     int dump_only,
+			     int removing,
+			     struct module_options **options,
+			     struct module_command **commands,
+			     struct module_alias **aliases,
+			     struct module_blacklist **blacklist)
 {
 	DIR *dir;
 	int ret = 0;
 
-	/* Reiser4 has file/directory duality: treat it as both. */
 	dir = opendir(filename);
 	if (dir) {
+		struct file_entry {
+			struct list_head node;
+			char name[];
+		};
+		LIST_HEAD(files_list);
+		struct file_entry *fe, *fe_tmp;
 		struct dirent *i;
-		while ((i = readdir(dir)) != NULL) {
-			if (!streq(i->d_name,".") && !streq(i->d_name,"..")
-			    && config_filter(i->d_name)) {
-				char sub[strlen(filename) + 1
-					 + strlen(i->d_name) + 1];
 
-				sprintf(sub, "%s/%s", filename, i->d_name);
-				if (!read_config(sub, name,
-						 dump_only, removing, options,
-						 commands, aliases, blacklist))
-					warn("Failed to open"
-					     " config file %s: %s\n",
-					     sub, strerror(errno));
-			}
+		/* sort files from directory into list */
+		while ((i = readdir(dir)) != NULL) {
+			size_t len;
+
+			if (i->d_name[0] == '.')
+				continue;
+			if (!config_filter(i->d_name))
+				continue;
+
+			len = strlen(i->d_name);
+			if (len < 6 ||
+			    (strcmp(&i->d_name[len-5], ".conf") != 0 &&
+			     strcmp(&i->d_name[len-6], ".alias") != 0))
+				warn("All config files need .conf: %s/%s, "
+				     "it will be ignored in a future release.\n",
+				     filename, i->d_name);
+			fe = malloc(sizeof(struct file_entry) + len + 1);
+			if (fe == NULL)
+				continue;
+			strcpy(fe->name, i->d_name);
+			list_for_each_entry(fe_tmp, &files_list, node)
+				if (strcmp(fe_tmp->name, fe->name) >= 0)
+					break;
+			list_add_tail(&fe->node, &fe_tmp->node);
 		}
 		closedir(dir);
+
+		/* parse list of files */
+		list_for_each_entry_safe(fe, fe_tmp, &files_list, node) {
+			char *cfgfile;
+
+			nofail_asprintf(&cfgfile, "%s/%s", filename, fe->name);
+			if (!parse_config_file(cfgfile, name,
+					       dump_only, removing,
+					       options, commands,
+					       aliases, blacklist))
+				warn("Failed to open config file "
+				     "%s: %s\n", fe->name, strerror(errno));
+			free(cfgfile);
+			list_del(&fe->node);
+			free(fe);
+		}
+
 		ret = 1;
+	} else {
+		if (parse_config_file(filename, name, dump_only, removing,
+				      options, commands, aliases, blacklist))
+			ret = 1;
 	}
-
-	if (read_config_file(filename, name, dump_only, removing,
-			     options, commands, aliases, blacklist))
-		ret = 1;
-
 	return ret;
 }
 
 /* Read binary index file containing aliases only */
-/* fallback to legacy aliases file as necessary */
-static int read_config_file_bin(const char *filename,
-			    const char *name,
-			    int dump_only,
-			    int removing,
-			    struct module_options **options,
-			    struct module_command **commands,
-			    struct module_alias **aliases,
-			    struct module_blacklist **blacklist)
+static void parse_toplevel_config(const char *filename,
+				  const char *name,
+				  int dump_only,
+				  int removing,
+				  struct module_options **options,
+				  struct module_command **commands,
+				  struct module_alias **aliases,
+				  struct module_blacklist **blacklist)
 {
-	struct index_value *realname;
-	char *binfile;
-	FILE *cfile;
-
-	nofail_asprintf(&binfile, "%s.bin", filename);
-	cfile = fopen(binfile, "r");
-	if (!cfile) {
-		free(binfile);
-		
-		return read_config_file(filename, name, dump_only, removing,
-					options, commands, aliases, blacklist);
-	}
-
-	if (dump_only) {
-		index_dump(cfile, stdout, "alias ");
-		free(binfile);
-		fclose(cfile);
-		return 1;
-	}
-	
-	realname = index_searchwild(cfile, name);
-	while(realname) {
-		struct index_value *next = realname->next;
-		*aliases = add_alias(realname->value, *aliases);
-		
-		free(realname);
-		realname = next;
-	}
-	
-	free(binfile);
-	fclose(cfile);
-	return 1;
-}
-
-static const char *default_configs[] = 
-{
-	"/etc/modprobe.conf",
-	"/etc/modprobe.d",
-};
-
-static void read_toplevel_config(const char *filename,
-				 const char *name,
-				 int dump_only,
-				 int removing,
-				 struct module_options **options,
-				 struct module_command **commands,
-				 struct module_alias **aliases,
-				 struct module_blacklist **blacklist)
-{
-	unsigned int i;
-
 	if (filename) {
-		if (!read_config(filename, name, dump_only, removing,
-				 options, commands, aliases, blacklist))
+		if (!parse_config_scan(filename, name, dump_only, removing,
+				       options, commands, aliases, blacklist))
 			fatal("Failed to open config file %s: %s\n",
 			      filename, strerror(errno));
 		return;
 	}
 
-	/* Try defaults. */
-	for (i = 0; i < ARRAY_SIZE(default_configs); i++) {
-		read_config(default_configs[i], name, dump_only, removing,
-				options, commands, aliases, blacklist);
-	}
+	/* deprecated config file */
+	if (parse_config_file("/etc/modprobe.conf", name, dump_only, removing,
+			      options, commands, aliases, blacklist) > 0)
+		warn("Deprecated config file /etc/modprobe.conf, "
+		      "all config files belong into /etc/modprobe.d/.\n");
+
+	/* default config */
+	parse_config_scan("/etc/modprobe.d", name, dump_only, removing,
+			  options, commands, aliases, blacklist);
 }
 
 /* Read possible module arguments from the kernel command line. */
-static int read_kcmdline(int dump_only, struct module_options **options)
+static int parse_kcmdline(int dump_only, struct module_options **options)
 {
 	char *line;
 	unsigned int linenum = 0;
@@ -1467,6 +1384,9 @@ static char **merge_args(char *args, char *argv[], int *argc)
 		newargs[num_env] = arg;
 	}
 
+	if (!newargs)
+		return argv;
+
 	/* Append commandline args */
 	newargs[0] = argv[0];
 	for (i = 1; i <= *argc; i++)
@@ -1497,7 +1417,7 @@ static char *gather_options(char *argv[])
 	return optstring;
 }
 
-static void handle_module(const char *modname,
+static int handle_module(const char *modname,
 			  struct list_head *todo_list,
 			  const char *newname,
 			  int remove,
@@ -1505,22 +1425,15 @@ static void handle_module(const char *modname,
 			  int first_time,
 			  errfn_t error,
 			  int dry_run,
-			  int verbose,
 			  struct module_options *modoptions,
 			  struct module_command *commands,
 			  int ignore_commands,
 			  int ignore_proc,
 			  int strip_vermagic,
 			  int strip_modversion,
-			  int unknown_silent,
 			  const char *cmdline_opts,
 			  int flags)
 {
-	struct stat finfo;
-
-	if (stat("/sys/module", &finfo) < 0)
-		fatal("/sys is not mounted.\n");
-
 	if (list_empty(todo_list)) {
 		const char *command;
 
@@ -1528,25 +1441,26 @@ static void handle_module(const char *modname,
 		   handle case where the first is completely bogus. */
 		command = find_command(modname, commands);
 		if (command && !ignore_commands) {
-			do_command(modname, command, verbose, dry_run, error,
+			do_command(modname, command, dry_run, error,
 				   remove ? "remove":"install", cmdline_opts);
-			return;
+			return 0;
 		}
 
-		if (unknown_silent)
-			exit(1);
-		error("Module %s not found.\n", modname);
-		return;
+		if (!quiet)
+			error("Module %s not found.\n", modname);
+		return 1;
 	}
 
 	if (remove)
-		rmmod(todo_list, newname, first_time, error, dry_run, verbose,
+		rmmod(todo_list, newname, first_time, error, dry_run,
 		      commands, ignore_commands, 0, cmdline_opts, flags);
 	else
 		insmod(todo_list, NOFAIL(strdup(options)), newname,
-		       first_time, error, dry_run, verbose, modoptions,
+		       first_time, error, dry_run, modoptions,
 		       commands, ignore_commands, ignore_proc, strip_vermagic,
 		       strip_modversion, cmdline_opts);
+
+	return 0;
 }
 
 static struct option options[] = { { "verbose", 0, NULL, 'v' },
@@ -1556,7 +1470,6 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "remove", 0, NULL, 'r' },
 				   { "wait", 0, NULL, 'w' },
 				   { "showconfig", 0, NULL, 'c' },
-				   { "autoclean", 0, NULL, 'k' },
 				   { "quiet", 0, NULL, 'q' },
 				   { "show", 0, NULL, 'n' },
 				   { "dry-run", 0, NULL, 'n' },
@@ -1585,8 +1498,6 @@ int main(int argc, char *argv[])
 	int dump_only = 0;
 	int dry_run = 0;
 	int remove = 0;
-	int verbose = 0;
-	int unknown_silent = 0;
 	int list_only = 0;
 	int all = 0;
 	int ignore_commands = 0;
@@ -1605,12 +1516,13 @@ int main(int argc, char *argv[])
 	char *aliasfilename, *symfilename;
 	errfn_t error = fatal;
 	int flags = O_NONBLOCK|O_EXCL;
+	int was_error = 0;
 
 	/* Prepend options from environment. */
 	argv = merge_args(getenv("MODPROBE_OPTIONS"), argv, &argc);
 
 	uname(&buf);
-	while ((opt = getopt_long(argc, argv, "vVC:o:rknqQsclt:aifbwd:", options, NULL)) != -1){
+	while ((opt = getopt_long(argc, argv, "vVC:o:rnqQsclt:aifbwd:", options, NULL)) != -1){
 		switch (opt) {
 		case 'v':
 			add_to_env_var("-v");
@@ -1629,7 +1541,7 @@ int main(int argc, char *argv[])
 			add_to_env_var(config);
 			break;
 		case 'q':
-			unknown_silent = 1;
+			quiet = 1;
 			add_to_env_var("-q");
 			break;
 		case 'D':
@@ -1657,9 +1569,6 @@ int main(int argc, char *argv[])
 			all = 1;
 			error = warn;
 			break;
-		case 'k':
-			/* FIXME: This should actually do something */
-			break;
 		case 'n':
 			dry_run = 1;
 			break;
@@ -1731,20 +1640,14 @@ int main(int argc, char *argv[])
 		struct module_alias *aliases = NULL;
 		struct module_blacklist *blacklist = NULL;
 
-		read_toplevel_config(config, "", 1, 0,
-			     &modoptions, &commands, &aliases, &blacklist);
-		read_kcmdline(1, &modoptions);
-		if (use_binary_indexes) {
-			read_config_file_bin(aliasfilename, "", 1, 0,
-			     &modoptions, &commands, &aliases, &blacklist);
-			read_config_file_bin(symfilename, "", 1, 0,
-			     &modoptions, &commands, &aliases, &blacklist);
-		} else {
-			read_config(aliasfilename, "", 1, 0,
-			     &modoptions, &commands, &aliases, &blacklist);
-			read_config(symfilename, "", 1, 0,
-			     &modoptions, &commands, &aliases, &blacklist);
-		}
+		parse_toplevel_config(config, "", 1, 0, &modoptions,
+				      &commands, &aliases, &blacklist);
+		/* Read module options from kernel command line */
+		parse_kcmdline(1, &modoptions);
+		parse_config_file(aliasfilename, "", 1, 0, &modoptions,
+				  &commands, &aliases, &blacklist);
+		parse_config_file(symfilename, "", 1, 0, &modoptions,
+				  &commands, &aliases, &blacklist);
 		exit(0);
 	}
 
@@ -1774,23 +1677,17 @@ int main(int argc, char *argv[])
 		underscores(modulearg);
 
 		/* Returns the resolved alias, options */
-		read_toplevel_config(config, modulearg, 0,
+		parse_toplevel_config(config, modulearg, 0,
 		     remove, &modoptions, &commands, &aliases, &blacklist);
 
 		/* Read module options from kernel command line */
-		read_kcmdline(0, &modoptions);
+		parse_kcmdline(0, &modoptions);
 
 		/* No luck?  Try symbol names, if starts with symbol:. */
-		if (!aliases
-		    && strncmp(modulearg, "symbol:", strlen("symbol:")) == 0) {
-			if (use_binary_indexes)
-				read_config_file_bin(symfilename, modulearg, 0,
-					remove, &modoptions, &commands,
-					&aliases, &blacklist);
-			else
-				read_config(symfilename, modulearg, 0,
-					remove, &modoptions, &commands,
-					&aliases, &blacklist);
+		if (!aliases && strstarts(modulearg, "symbol:")) {
+			parse_config_file(symfilename, modulearg, 0,
+					  remove, &modoptions, &commands,
+					  &aliases, &blacklist);
 		}
 		if (!aliases) {
 			if(!strchr(modulearg, ':'))
@@ -1800,16 +1697,10 @@ int main(int argc, char *argv[])
 			if (list_empty(&list)
 			    && !find_command(modulearg, commands))
 			{
-				if (use_binary_indexes)
-					read_config_file_bin(aliasfilename,
-						modulearg, 0, remove,
-						&modoptions, &commands,
-						&aliases, &blacklist);
-				else
-					read_config(aliasfilename,
-						modulearg, 0, remove,
-						&modoptions, &commands,
-						&aliases, &blacklist);
+				read_aliases_file(aliasfilename,
+						  modulearg, 0, remove,
+						  &modoptions, &commands,
+						  &aliases, &blacklist);
 			}
 		}
 
@@ -1827,14 +1718,15 @@ int main(int argc, char *argv[])
 							 opts, modoptions);
 
 				read_depends(dirname, aliases->module, &list);
-				handle_module(aliases->module, &list, newname,
-					      remove, opts, first_time, err,
-					      dry_run, verbose, modoptions,
+				if (handle_module(aliases->module, &list,
+					      newname, remove, opts,
+					      first_time, err,
+					      dry_run, modoptions,
 					      commands, ignore_commands,
 					      ignore_proc, strip_vermagic,
 					      strip_modversion,
-					      unknown_silent,
-					      optstring, flags);
+					      optstring, flags))
+					was_error = 1;
 
 				aliases = aliases->next;
 				INIT_LIST_HEAD(&list);
@@ -1844,12 +1736,13 @@ int main(int argc, char *argv[])
 			    && find_blacklist(modulearg, blacklist))
 				continue;
 
-			handle_module(modulearg, &list, newname, remove,
+			if (handle_module(modulearg, &list, newname, remove,
 				      optstring, first_time, error, dry_run,
-				      verbose, modoptions, commands,
+				      modoptions, commands,
 				      ignore_commands, ignore_proc,
 				      strip_vermagic, strip_modversion,
-				      unknown_silent, optstring, flags);
+				      optstring, flags))
+				was_error = 1;
 		}
 	}
 	if (logging)
@@ -1859,6 +1752,9 @@ int main(int argc, char *argv[])
 	free(aliasfilename);
 	free(symfilename);
 	free(optstring);
-	
-	return 0;
+
+	if (was_error)
+		exit(1);
+	else
+		exit(0);
 }
